####################################################
# Researcher: Sara Schwarz
# Advisor: Dr. Jose Ortiz
#
# Program Objective:
#
#In this work we seek to develop algorithms (brute-force approach, TRW, etc) to detect network and port
#scanners on large-scale networks traffc such as those of Internet Service Providers,
#Big Data research centers, and Science DMZ networks implemented in research institutions using network
#flows. This specific program will follow a brute-force approach algorithm, of reading the network flows
#and recording for each source ip the destination ips and the different destination ports it connected to.
#Later, the number of connected ports will be compared with a threshold to classify the source ips as
#either suspicious scanners or not.
###############################################################

from silk import *



#input => list of files where each file has a list of network flows
#output => threeway-hash table of the form {sip:{dip:{dport:total # of each dport}}}
def verify_type(filename):
    #print "Files", len(filename)

    count_flow = 0
    dportHash = {} #will contain amount of dport per dip per each sip
    #iterates through each file that has a list of network flows
    for file in filename:

        #iterates through each network flow, and reads it using silkfile_open, SiLks Flow Repository function
        for rec in silkfile_open(file, READ):
            count_flow +=1
            sip = str(rec.sip)
            dip = str(rec.dip)
            dport = rec.dport
            #For our current research, we are limiting to only verifying IPv4 network flows
            if (':' in sip):
                continue
            #using the hash table structure, keeps record of all the times each
            #destination port for each destination ip was connected to by the same source ip
            else:
                if sip in dportHash:
                    if dip in dportHash[sip]:
                        if dport in dportHash[sip][dip]:
                            dportHash[sip][dip][dport] += 1
                        else:
                            dportHash[sip][dip][dport] = 1
                    else:
                        dportHash[sip][dip] = {dport : 1}
                else:
                    dportHash[sip] = { dip: {dport: 1} }


    #print "flows", count_flow
    return dportHash


#MAIN
#for accesing the network flows during a specific interval of time
startDate = "2018/08/1"
endDate = "2018/10/15"
otherHash = {}#hash table for suspicious source ips with destination ips
counter = 0
threshold = 100
#object for retrieving files from Silk data store. Specified the interval of time, the silk configuration file location, and the silk data location
files1 = FGlob(classname="all", type="all", start_date=startDate, end_date=endDate, site_config_file="/etc/silk/conf-v9/silk.conf", data_rootdir="/home/scratch/flow/rwflowpack/")
files = [x for x in files1]
#****************************************
#  Verifying the source ips
#****************************************
flowHash = verify_type()
#Will iterate through the hash of sips and dports
#Verify whether the number of different dports per desitnation ip, connected to by the same source ip
#is greater than our threshold. If it is, the source ip will be added to the hash suspicious sips.

for sips in flowHash:
    for dips, dports in flowHash[sips].items():
        if len(dports) >= threshold:
            if sips in otherHash:
                otherHash[sips][dips] = dports
            else:
                otherHash[sips] = {dips: dports}

for dips, dports in otherHash.items():
    counter +=1
    #prints the total number of suspicious source ips.

print counter
