####################################################
# Researcher: Sara Schwarz
# Advisor: Dr. Jose Ortiz
# VERSION #2
# Program Objective:
#
#In this work we seek to develop algorithms (brute-force approach, TRW, etc) to detect network and port
#scanners on large-scale networks traffc such as those of Internet Service Providers,
#Big Data research centers, and Science DMZ networks implemented in research institutions using network
#flows. This specific program will follow a brute-force approach algorithm, of reading the network flows
#and recording for each source ip the destination ips and the different destination ports it connected to.
#Later, the number of connected ports will be compared with a threshold to classify the source ips as
#either suspicious scanners or not.
#To run this algorithm we will be using high-performance methods for computing such as Map and Reduce,
#specifically Python's Pool Class Library.
###############################################################


from silk import *
import multiprocessing as mp


#input => list of files where each file has a list of network flows
#output => threeway-hash table of the form {sip:{dip:{dport:total # of each dport}}}
def verify_type(filename):
    #print "Files", len(filename)

    count_flow = 0
    dportHash = {} #will contain amount of dport per dip per each sip
    #iterates through each file that has a list of network flows
    for file in filename:

        #iterates through each network flow, and reads it using silkfile_open, SiLks Flow Repository function
        for rec in silkfile_open(file, READ):
            count_flow +=1
            sip = str(rec.sip)
            dip = str(rec.dip)
            dport = rec.dport
            #For our current research, we are limiting to only verifying IPv4 network flows
            if (':' in sip):
                continue
            #using the hash table structure, keeps record of all the times each
            #destination port for each destination ip was connected to by the same source ip
            else:
                if sip in dportHash:
                    if dip in dportHash[sip]:
                        if dport in dportHash[sip][dip]:
                            dportHash[sip][dip][dport] += 1
                        else:
                            dportHash[sip][dip][dport] = 1
                    else:
                        dportHash[sip][dip] = {dport : 1}
                else:
                    dportHash[sip] = { dip: {dport: 1} }


    #print "flows", count_flow
    return dportHash

#input => list of hash tables of the form {sip:{dip:{dport:total # of each dport}}}
#output => hash table of the form {sip:{dip:{dport:total # of each dport}}}
#Method => iterates through each hash
def join_hash(list):
    complete_hash ={} #will contain the hash table of all the hash tables once formed

    for i in list: #will iterate through each hash of the list
        #will iterate through each source ip key as sip with the hash table of the form {dip:{dport:#}} as the value
        for sip, hash in i.items():
            #verify if that source ip is already in the merged hash.
            if sip in complete_hash:
                #will iterate through each destination ip key as dip with the hash table of the form {dport:#} as the value
                for dip, dports in i[sip].items():
                    #verify if that destination ip is already in the hash of the sip's value in the merged hash
                    if dip in complete_hash[sip]:
                        #will iterate through each destination port key as number with the total # of times it was connected as the value
                        for number, value in dports.items():
                            #verify if that destination port is already in the hash of the dip's value in the hash of the sips'value in the merged hash
                            if number in complete_hash[sip]:
                                #Add the value to the dport's value in the merged hash
                                complete_hash[sip][dip][number] += value

                            else:
                                #Add the destination port with its value to the merged hash
                                complete_hash[sip][dip][number]= value
                    else:
                        #Add the destination ip with its hash table of dports to the merged hash
                        complete_hash[sip][dip]= dports
            else:
                #Add the source ip with its hash table of destination ip and dports to the merged hash
                complete_hash[sip]= hash
    return complete_hash


def main():
    #for accesing the network flows during a specific interval of time
    startDate = "2018/08/1"
    endDate = "2018/10/15"
    otherHash = {} #hash table for suspicious source ips with destination ips
    counter = 0
    threshold = 100
    #object for retrieving files from Silk data store. Specified the interval of time, the silk configuration file location, and the silk data location
    files = FGlob(classname="all", type="all", start_date=startDate, end_date=endDate, site_config_file="/etc/silk/conf-v9/silk.conf", data_rootdir="/home/scratch/flow/rwflowpack/")
#****************************************
# Using Pythons Pool Class for multiprocessing
#****************************************
    process_num = 2
    pool = mp.Pool(processes=process_num)
    #change files1 from FGlob object type to a list of silk data files
    files1 = [x for x in files]
    files_list = []
    #Send each process an equal amount of files
    blocksize = len(files1) / process_num
    for x in range(process_num):
        files_list.append(files1[0:blocksize])
        files1 = files1[blocksize:]
    #in the case that the number of files is not divisible by the number of processes
    for i in files1:
        files_list[files1.index(i)].append(i)
    #Using map from Pool, returns a list of a hash per process, being the hash the output of verify_type
    fileHash = pool.map(verify_type, files_list)
    #returns the hash of all hashes merged.
    flowHash = join_hash(fileHash)
    #print len(flowHash)
#****************************************

#****************************************
#  Verifying the source ips
#****************************************
    #Will iterate through the hash of sips and dports
    #Verify whether the number of different dports per desitnation ip, connected to by the same source ip
    #is greater than our threshold. If it is, the source ip will be added to the hash suspicious sips.
    for sips in flowHash:
        for dips, dports in flowHash[sips].items():
            if len(dports) >= threshold:
                if sips in otherHash:
                    otherHash[sips][dips] = dports
                else:
                    otherHash[sips] = {dips: dports}

    for dips, dports in otherHash.items():
        counter +=1
    #prints the total number of suspicious source ips.
    print counter


if __name__== "__main__":
  main()




  # def Filter(fd, sip, dip, sport, dport, dportHash):
  #
  # 	if dport > 1024 and (sport <= 1024 or (sport >= 8000 and sport < 9000)):
  # 		return
  # 	if sip in dportHash:
  # #       	if dip in dportHash[sip]["dips"]:
  # #        	        dportHash[sip]["dips"][dip] += 1
  # #		else:
  # #			dportHash[sip]["dips"][dip] = 1
  # 		if dport in dportHash[sip]["dports"]:
  #                	        dportHash[sip]["dports"][dport] += 1
  # 			#return
  #                	else:
  #                		dportHash[sip]["dports"][dport] = 1
  #      	else:
  # 		dportHash[sip] = {"dports": {}}
  # 		dportHash[sip]["dips"] = {}
  # 	#fd.write("%s:%s:%s:%s\n" % (sip, dip, sport, dport))
